###############################################################################
# Copyright (C) 2022 Habana Labs, Ltd. an Intel Company
# All Rights Reserved.
#
# Unauthorized copying of this file or any element(s) within it, via any medium
# is strictly prohibited.
# This file contains Habana Labs, Ltd. proprietary and confidential information
# and is subject to the confidentiality and license agreements under which it
# was provided.
#
###############################################################################
import tensorflow as tf
from habana_frameworks.tensorflow.library_loader import media_ops
from tensorflow.python.ops import gen_dataset_ops
from tensorflow.python.eager import function
from tensorflow.python.data.ops import dataset_ops
import gc
import atexit

class HabanaDataset(dataset_ops.DatasetSource):
    def __init__(self, output_shapes, output_types, pipeline):
        self._output_signature = tuple(map(lambda x: tf.TensorSpec(shape=x[0], dtype=x[1]),
                                           zip(output_shapes, output_types)))

        def glue_init():
            pipeline.iter_init()
            return self._proxy

        def glue_next():
            imgs, labels = pipeline.as_iterator()
            return imgs, labels

        @function.defun
        @tf.autograph.experimental.do_not_convert
        def init_func():
            return tf.py_function(glue_init, [], Tout=[tf.uint64])

        @function.defun(input_signature=[tf.TensorSpec(shape=(), dtype=tf.uint64)])
        @tf.autograph.experimental.do_not_convert
        def next_func(proxy):
            output_buffers = tf.py_function(
                glue_next, inp=[], Tout=[tf.uint64, tf.uint64])
            return media_ops.HabanaDatasetNext(proxy=proxy, tensor_addr=output_buffers,
                                               output_shapes=output_shapes, output_types=output_types)

        @function.defun(input_signature=[tf.TensorSpec(shape=(), dtype=tf.uint64)])
        @tf.autograph.experimental.do_not_convert
        def finalize_func(proxy):
            media_ops.HabanaDatasetFinalize(proxy=proxy)
            return tf.constant(0, tf.uint64)

        pipeline.build()
        self._proxy = media_ops.CreateMediaProxy()
        pipeline.set_proxy("TF_FW", int(self._proxy.numpy()))
        self._init_func = init_func._get_concrete_function_internal()
        self._next_func = next_func._get_concrete_function_internal()
        self._finalize_func = finalize_func._get_concrete_function_internal()
        self._init_captured_args = self._init_func.captured_inputs
        self._next_captured_args = self._next_func.captured_inputs
        self._finalize_captured_args = self._finalize_func.captured_inputs
        with tf.device("/device:HPU:0"):
            variant_tensor = gen_dataset_ops.generator_dataset(
                self._init_captured_args,
                self._next_captured_args,
                self._finalize_captured_args,
                init_func=self._init_func,
                next_func=self._next_func,
                finalize_func=self._finalize_func,
                output_shapes=output_shapes,
                output_types=output_types)

        # there is a hang happening on application exit, that has something to do with how tf.py_function & Python GIL
        # are interacting together. Many similar issues were reported to TensorFlow.
        # I.e. in https://github.com/tensorflow/tensorflow/issues/21277
        # Simple WA is to enforce gc.collect() at the end of application
        atexit.register(gc.collect)

        super(HabanaDataset, self).__init__(variant_tensor)

    @property
    def element_spec(self):
        return self._output_signature